/***************************************************************************
                          timemanager.cpp  -  description
                             -------------------
    begin                : Sun Dec 23 2001
    copyright            : (C) 2001, 2002 by Thomas Friedrichsmeier
    email                : Thomas.Friedrichsmeier@ruhr-uni-bochum.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "defines.h"
#include "timemanager.h"

#include "cdp.h"
#include <qstring.h>
#include <klocale.h>

// TARGET_SCORE * BASE_TI = maximum msecs a timer can have! Also a greate TARGET_SCORE increases resolution.
// TARGET_SCORE * BASE_TI should of course not exceed max_value for the used ints (int_fast32_t)
#define TARGET_SCORE 100000

TimeManager::TimeManager(int num_timers, int base_intervall){

	max_timers = num_timers;
	base_ti = base_intervall;

	timers = new Timer [max_timers+1];
	ids = new int [max_timers];

	for (int i = 0; i <= max_timers; i++) {
		timers[i].to = 0;
		timers[i].increase = 0;
		timers[i].score = 0;
		
		if (i < max_timers) {
			ids[i] = i;
		}
	}
	active_timers = 0;
	
	resume ();	
}

TimeManager::~TimeManager(){
	pause ();
	delete [] timers;
	delete [] ids;
}

void TimeManager::timerEvent(QTimerEvent *) {
	for (int i = 0; i < active_timers; i++) {
		int id = ids[i];						// look up timer
		timers[id].score += timers[id].increase;		// increase counter
		while (timers[id].score >= TARGET_SCORE) {			// reset and signal if score was reached (potentially several times)
			timers[id].score -= TARGET_SCORE;
			timers[id].to->timerTick (id);
		}
	}
	// do the same for the "last timer"
	timers[max_timers].score += timers[max_timers].increase;
	while (timers[max_timers].score >= TARGET_SCORE) {
		timers[max_timers].score -= TARGET_SCORE;
		timers[max_timers].to->timerTick (max_timers);
	}
}

/** Pauses ALL timers */
void TimeManager::pause (){
	QObject::killTimers ();      // merely kills the base-timer -> no more processing of timerEvent
}

/** Resumes all timers */
void TimeManager::resume (){
	pause ();			// just in case...
	QObject::startTimer (base_ti);
}

/** Starts a timer for TimedObject to with interval timer_period msec, returning id assigned to this timer.
	Returned ids are always in the range of 0 .. max_timers -1 */
int TimeManager::startTimer (TimedObject *to, int_fast32_t timer_period) {
	QString debug;

	int id = -1;
	if ((to != 0) && (timer_period > 0)) {
		if (active_timers < (max_timers-1)) {
			id = ids[active_timers++];
			timers[id].to = to;
			timers[id].increase = (TARGET_SCORE * base_ti) / timer_period;	//  * TARGET_SCORE is to give better accuracy (using ints)
		} else {
			cdp->debug_msg (i18n ("Timer Limit exceeded!"), Serious);
		}
	}
	return id;
}

/** Dirty hack: set a timer to be checked at the very end of the list. No sanity checks. Any previous timer at this
	place gets deleted without notice! */
int TimeManager::setLastTimer (TimedObject *to, int_fast32_t timer_period) {
	timers[max_timers].to = to;
	timers[max_timers].increase = (TARGET_SCORE * base_ti) / timer_period;

	return max_timers;
}

/** Kills a timer by its id */
void TimeManager::killTimer (int id) {
	int i = 0;
	while ((i < active_timers) && (ids[i] != id))	{		// find position of "id" in the list of used ids
		i++;
	}
	if (i < active_timers) {									// check whether "id" was really in use (does NOT check, whether timer belongs to caller)
		ids[i] = ids[active_timers-1];				// swap id to be deleted with "rightmost" id in list of ids --> see documentation on ids in header-file
		ids[active_timers-1] = id;
		active_timers--;

		timers[id].to = 0;								      // reset timer for next use
		timers[id].increase = timers[id].score = 0;
	}
}

/** Kills all timers for a specified object (slower) */
void TimeManager::killTimers (TimedObject *to) {
	for (int i = 0; i < max_timers; i++) {
		if (timers[i].to == to) {
			this->killTimer (i);		// avoid confusion with QObject::killTimer (int)
		}
	}
}

/** Kills the special last timer */
void TimeManager::killLastTimer () {
	timers[max_timers].to = 0;
	timers[max_timers].increase = timers[max_timers].score = 0;
}
